#!/usr/bin/env python

"""
Operator support.

Copyright (C) 2010, 2013 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

def binary_op(a, b, left_accessor, right_accessor):

    """
    A single parameterised function providing the binary operator mechanism for
    arguments 'a' and 'b' using accessors given as 'left_accessor' and
    'right_accessor' which provide the methods for the operands.
    """

    # First, try and get a method for the left argument, and then call it with
    # the right argument.

    try:
        fn = left_accessor(a)
    except AttributeError:
        pass
    else:
        result = fn(b)
        if result is not NotImplemented:
            return result

    # Otherwise, try and get a method for the right argument, and then call it
    # with the left argument.

    try:
        fn = right_accessor(b)
    except AttributeError:
        pass
    else:
        result = fn(a)
        if result is not NotImplemented:
            return result

    # Where no methods were available, or if neither method could support the
    # operation, raise an exception.

    raise TypeError

def unary_op(a, accessor):

    """
    A single parameterised function providing the unary operator mechanism for
    the argument 'a' using the given 'accessor' to provide the method for the
    operand.
    """

    # First, try and get a method for the argument, and then call it.

    try:
        fn = accessor(a)
    except AttributeError:
        pass
    else:
        result = fn()
        if result is not NotImplemented:
            return result

    # Where no method was available, or if the method could not support the
    # operation, raise an exception.

    raise TypeError

def augassign(a, b, augmented_accessor, left_accessor, right_accessor):

    """
    A single parameterised function providing the augmented assignment mechanism
    for arguments 'a' and 'b' either using 'augmented_accessor' (directly
    affecting 'a') or using 'left_accessor' and 'right_accessor' (conventional
    operator method accessors).

    The result of the assignment is returned.
    """

    # First, try and get a method that directly affects the assignment target.

    try:
        fn = augmented_accessor(a)
    except AttributeError:
        pass
    else:
        result = fn(b)
        if result is not NotImplemented:
            return result

    # Otherwise, attempt a conventional binary operation.

    return binary_op(a, b, left_accessor, right_accessor)

# These functions defer method lookup by wrapping the attribute access in
# lambda functions. Thus, the appropriate methods are defined locally, but no
# attempt to obtain them is made until the generic function is called.

# NOTE: The compiler should make it possible for the following functions to call
# NOTE: the generic operator implementations with no additional call overhead.

# Binary operator functions.

def add(a, b):
    return binary_op(a, b, lambda a: a.__add__, lambda b: b.__radd__)

def and_(a, b):
    return binary_op(a, b, lambda a: a.__and__, lambda b: b.__rand__)

def contains(a, b):
    return b in a

def div(a, b):
    return binary_op(a, b, lambda a: a.__div__, lambda b: b.__rdiv__)

def floordiv(a, b):
    return binary_op(a, b, lambda a: a.__floordiv__, lambda b: b.__rfloordiv__)

def in_(a, b):
    return a.__contains__(b)

def not_in(a, b):
    return not a.__contains__(b)

def lshift(a, b):
    return binary_op(a, b, lambda a: a.__lshift__, lambda b: b.__rlshift__)

def mod(a, b):
    return binary_op(a, b, lambda a: a.__mod__, lambda b: b.__rmod__)

def mul(a, b):
    return binary_op(a, b, lambda a: a.__mul__, lambda b: b.__rmul__)

def or_(a, b):
    return binary_op(a, b, lambda a: a.__or__, lambda b: b.__ror__)

def pow(a, b):
    return binary_op(a, b, lambda a: a.__pow__, lambda b: b.__rpow__)

def rshift(a, b):
    return binary_op(a, b, lambda a: a.__rshift__, lambda b: b.__rrshift__)

def sub(a, b):
    return binary_op(a, b, lambda a: a.__sub__, lambda b: b.__rsub__)

def xor(a, b):
    return binary_op(a, b, lambda a: a.__xor__, lambda b: b.__rxor__)

# Unary operator functions.

def invert(a):
    return unary_op(a, lambda a: a.__invert__)

def neg(a):
    return unary_op(a, lambda a: a.__neg__)

def not_(a):
    return not a

def pos(a):
    return unary_op(a, lambda a: a.__pos__)

# Augmented assignment functions.

def iadd(a, b):
    return augassign(a, b, lambda a: a.__iadd__, lambda a: a.__add__, lambda b: b.__radd__)

def iand_(a, b):
    return augassign(a, b, lambda a: a.__iand__, lambda a: a.__and__, lambda b: b.__rand__)

def idiv(a, b):
    return augassign(a, b, lambda a: a.__idiv__, lambda a: a.__div__, lambda b: b.__rdiv__)

def ifloordiv(a, b):
    return augassign(a, b, lambda a: a.__ifloordiv__, lambda a: a.__floordiv__, lambda b: b.__rfloordiv__)

def ilshift(a, b):
    return augassign(a, b, lambda a: a.__ilshift__, lambda a: a.__lshift__, lambda b: b.__rlshift__)

def imod(a, b):
    return augassign(a, b, lambda a: a.__imod__, lambda a: a.__mod__, lambda b: b.__rmod__)

def imul(a, b):
    return augassign(a, b, lambda a: a.__imul__, lambda a: a.__mul__, lambda b: b.__rmul__)

def ior_(a, b):
    return augassign(a, b, lambda a: a.__ior__, lambda a: a.__or__, lambda b: b.__ror__)

def ipow(a, b):
    return augassign(a, b, lambda a: a.__ipow__, lambda a: a.__pow__, lambda b: b.__rpow__)

def irshift(a, b):
    return augassign(a, b, lambda a: a.__irshift__, lambda a: a.__rshift__, lambda b: b.__rrshift__)

def isub(a, b):
    return augassign(a, b, lambda a: a.__isub__, lambda a: a.__sub__, lambda b: b.__rsub__)

def ixor(a, b):
    return augassign(a, b, lambda a: a.__ixor__, lambda a: a.__xor__, lambda b: b.__rxor__)

# Comparison functions.

def eq(a, b):
    return binary_op(a, b, lambda a: a.__eq__, lambda b: b.__eq__)

def ge(a, b):
    return binary_op(a, b, lambda a: a.__ge__, lambda b: b.__le__)

def gt(a, b):
    return binary_op(a, b, lambda a: a.__gt__, lambda b: b.__lt__)

def le(a, b):
    return binary_op(a, b, lambda a: a.__le__, lambda b: b.__ge__)

def lt(a, b):
    return binary_op(a, b, lambda a: a.__lt__, lambda b: b.__gt__)

def ne(a, b):
    return binary_op(a, b, lambda a: a.__ne__, lambda b: b.__ne__)

# Access and slicing functions.

def getitem(a, b):
    return a.__getitem__(b)

def setitem(a, b, c):
    a.__setitem__(b, c)

# NOTE: Should be able to optimise temporary instance allocations for slices.

def getslice(a, b, c):
    return a.__getitem__(slice(b, c))

def setslice(a, b, c, d):
    a.__setitem__(slice(b, c), d)

# vim: tabstop=4 expandtab shiftwidth=4
